Promise = require('bluebird')
Http = require('http')
Request = require('request')
Url = require('url')

Logger = require('./logger')
PackageJson = require('./package.json')
Utils = require('./utils')


API_CONFIG =
  allowUnverified: false
  key: process.env.FAAS_API_KEY
  secret: process.env.FAAS_API_SECRET
  hostname: 'api.faas.io'

debug = Logger('api') # Local logger.


exports.config = (config) ->
  if config
    apiConf = {}
    apiConf.key = config.apiKey if config.apiKey
    apiConf.secret = config.apiSecret if config.apiSecret
    exports.apiConfig(apiConf)


exports.apiConfig = (config) ->
  Utils.extend(API_CONFIG, config)
  Utils.shallowCopy(API_CONFIG)


###
# Interface for performing API calls.
###
exports.api = (options, callback) ->
  debug('options', options)
  callback = debug.wrap(callback)

  urlOpts = options.url
  # Default the credentials.
  query = { apiKey: API_CONFIG.key, apiSecret: API_CONFIG.secret }
  if urlOpts.query
    Utils.extend(query, urlOpts.query)
  query.npmVer = PackageJson.version
  urlOpts.query = query
  urlOpts.hostname = API_CONFIG.hostname
  urlOpts.protocol = 'https:'

  # Be clear about configuration errors.
  cfgErrors = []
  cfgErrors.push('blank api hostname') unless urlOpts.hostname
  cfgErrors.push('missing api key') unless urlOpts.query.apiKey
  cfgErrors.push('missing api secret') unless urlOpts.query.apiSecret
  if cfgErrors.length > 0
    debug('ERROR urlOpts', urlOpts)
    callback('Invalid configuration options: '+cfgErrors.join('; '))
    return

  reqOpts =
    uri: Url.format(urlOpts)
    method: options.method || 'GET'
    json: options.data || true
    encoding: 'utf8'
    strictSSL: if API_CONFIG.allowUnverified then false else true # Disable SSL certificate checking.

  debug('reqOpts', reqOpts)

  Request reqOpts, (err, res, body) ->
    debug('response', res && res.statusCode, res && res.headers, body)
    if !err && res && (res.statusCode < 200 || res.statusCode >= 300)
      err = body && body.error || (res.statusCode+": "+Http.STATUS_CODES[res.statusCode])+': '+JSON.stringify(body)
    if body && body.data
      body = body.data
    callback(err, body)

exports.apiQ = (options) ->
  Promise.fromCallback (callback) ->
    exports.api(options, callback)
