Stream = require('stream')

Request = require('request')
SocketIoStream = require('socket.io-stream')

Base = require('../base')
Logger = require('./logger')


debug = Base.logger('http_proxy')

PROXY_HTTP_HOST_USE_PROXY = 'USE_PROXY'

upstreamUrl = null
isUpstreamHttps = null


exports.setupHttpProxy = (upstream) ->
  upstreamUrl = upstream || Base.env.IX_HTTP_PROXY
  upstreamUrl = upstreamUrl?.trim().replace(/\/$/, '')
  throw new Error("No upstream server defined.") unless upstreamUrl
  debug('setupHttpProxy upstreamUrl', upstreamUrl)
  isUpstreamHttps = upstreamUrl.slice(0,6) == 'https:'

exports.onHttpProxy = (faasReq, cb_proxyResponse) ->
  Logger.logHttpRequest(faasReq)
  cb_proxyResponse = debug.wrap(cb_proxyResponse)
  url = upstreamUrl + faasReq.url
  fwdOptions =
    url: url
    headers: faasReq.headers
    method: faasReq.method
    followRedirect: false
  httpHost = Base.env.IX_PROXY_HTTP_HOST
  if isUpstreamHttps || httpHost == PROXY_HTTP_HOST_USE_PROXY
    delete fwdOptions.headers.host
  if httpHost && httpHost != PROXY_HTTP_HOST_USE_PROXY
    fwdOptions.headers.host = httpHost
  debug('request', fwdOptions)
  fwdReq = null
  requestStream = faasReq.requestStream
  proxyResponse = {}
  reqCounter = new Stream.Transform()
  reqCounter.numBytes = 0
  reqCounter._transform = (chunk, encoding, callback) ->
    debug('request chunk', encoding, chunk)
    console.log(chunk.toString())
    reqCounter.numBytes += chunk.length
    callback(null, chunk)
  try
    fwdReq = Request(fwdOptions)
  catch e
    # Error: Invalid protocol: localhost:
    proxyResponse.statusCode = 500
    proxyResponse.error = e.toString()
    Logger.logHttpResponseError(faasReq, proxyResponse)
    cb_proxyResponse(proxyResponse)
    requestStream.on('data', (data) -> debug('fwdReq.error data', data) )
    # requestStream.end(e.toString())
    return
  responseStarted = false
  responseStream = SocketIoStream.createStream()
  proxyResponse.responseStream = responseStream
  faasReq.responseBytes = 0
  respCounter = new Stream.Transform()
  respCounter._transform = (chunk, encoding, callback) ->
    debug('response chunk', encoding, chunk)
    faasReq.responseBytes += chunk.length
    callback(null, chunk)
  fwdReq.on 'error', (err) ->
    debug('fwdReq.error', err, proxyResponse)
    if responseStarted
      console.error('duplicate ERROR response attempted for '+url)
      responseStream.emit('error', err)
    else
      responseStarted = true
      proxyResponse.error = err.toString()
      cb_proxyResponse(proxyResponse)
    try
      fwdReq.abort()
      fwdReq.destroy()
      requestStream.unpipe()
      requestStream.on('data', (data) -> debug('fwdReq.error data', data) )
      responseStream.end()
    catch e
      console.error(e)
    Logger.logHttpResponseError(faasReq, proxyResponse)

  fwdReq.on 'response', (response) ->
    Logger.logHttpResponseStart(faasReq, response)
    debug('fwdReq.response', response.headers, response.statusCode, response.statusMessage, proxyResponse)
    if responseStarted
      console.error('duplicate SUCCESS response attempted for '+url)
      return
    responseStarted = true
    proxyResponse.lag = faasReq.ts.responseStart - faasReq.ts.requestStart
    proxyResponse.headers = response.headers
    proxyResponse.statusCode = response.statusCode
    proxyResponse.statusMessage = response.statusMessage
    cb_proxyResponse(proxyResponse)
    fwdReq.pipe(respCounter).pipe(responseStream)

  fwdReq.on 'end', ->
    Logger.logHttpResponseEnd(faasReq)

  requestStream.pipe(reqCounter).pipe(fwdReq)
